#include "AssemblyGenerator.h"

using namespace zp;
using namespace std;

#include <iostream>


#ifdef __APPLE__
const char* LabelNames::main = "_main";
const char* LabelNames::getchar = "_getchar";
const char* LabelNames::putchar = "_putchar";
const char* LabelNames::calloc = "_calloc";
const char* LabelNames::free = "_free";
#else
const char* LabelNames::main = "main";
const char* LabelNames::getchar = "getchar";
const char* LabelNames::putchar = "putchar";
const char* LabelNames::calloc = "calloc";
const char* LabelNames::free = "free";
#endif


const char* AssemblyGenerator::pointerRegister = "rbx";
const char* AssemblyGenerator::dataRegister = "r12";

#ifdef _WIN32
const char* AssemblyGenerator::argumentRegister1 = "rcx";
const char* AssemblyGenerator::argumentRegister2 = "rdx";
#else
const char* AssemblyGenerator::argumentRegister1 = "rdi";
const char* AssemblyGenerator::argumentRegister2 = "rsi";
#endif


AssemblyGenerator::AssemblyGenerator(std::ostream& out) :
    out{ out }
{
    write(".globl ", LabelNames::main);
    write(LabelNames::main, ":");
    write("subq $40, %rsp");
    write("movq $131072, %", argumentRegister1);
    write("movq $1, %", argumentRegister2);
    write("callq ", LabelNames::calloc);
    write("addq $65536, %rax");
    write("movq %rax, %", dataRegister);
    write("xorq %", pointerRegister, ", %", pointerRegister);
    write("movq %", dataRegister, ", %", pointerRegister);
    write();
    write("# begin of bf-program");
    write();
}


void AssemblyGenerator::finish(void)
{
    write();
    write("# end of bf-program");
    write();
    write("movq %", dataRegister, ", %", argumentRegister1);
    write("subq $65536, %", argumentRegister1);
    write("callq ", LabelNames::free);
    write("xorq %rax, %rax");
    write("addq $40, %rsp");
    write("retq");
}


std::string AssemblyGenerator::generateLabel(void)
{
    return "L" + std::to_string(labelcount++);
}


std::string AssemblyGenerator::generateOffset(const std::string& offset,
    const std::string& base)
{
    return "(" + offset + ", " + base + ")";
}


std::string AssemblyGenerator::ptr(void)
{
    //return generateOffset("%"s + pointerRegister, "%"s + dataRegister);
    return "(%"s + pointerRegister + ")";
}

void AssemblyGenerator::visitBlockInstruction(const BlockInstruction&)
{
}


void AssemblyGenerator::visitUnionBlockInstruction(const UnionBlockInstruction& ubi)
{
    for (auto& a : ubi.content) {
        a->accept(*this);
    }
}


void AssemblyGenerator::visitIOInstruction(const IOInstruction& ioi)
{
    if (ioi.isRead) {
        write("callq ", LabelNames::getchar);
        write("movb %al, ", ptr());
    }
    else {
        write("movq ", ptr(), ", %", argumentRegister1);
        write("andq $255, %", argumentRegister1);
        write("callq ", LabelNames::putchar);
    }
}


void AssemblyGenerator::visitSimpleBlockInstruction(const SimpleBlockInstruction& sbi)
{
    for (const auto [index, value] : sbi.added) {
        write("addb $", int(value), ", ", index,
            ptr());
    }
    if (sbi.pointerMoved != 0)
        write("addq $", sbi.pointerMoved, ", %", pointerRegister);
    //write("andq $65535, %", pointerRegister);
}


void AssemblyGenerator::visitLinearLoopInstruction(const LinearLoopInstruction& lli)
{
    if (lli.increment == 255) {
        if (!lli.factors.empty())
            write("movzbl ", ptr(), ", %eax");
        for (const auto [index, value] : lli.factors) {
            compileMult(index, value);
        }
        write("movb $0, ", ptr());
    }
    else throw "very fatal error";
}


void AssemblyGenerator::visitLoopInstruction(const LoopInstruction& li)
{
    std::string label1 = generateLabel();
    std::string label2 = generateLabel();
    write(label1, ":");
    write("movb ", ptr(), ", %al");
    write("testb %al, %al");
    write("jz ", label2);
    visitUnionBlockInstruction(li);
    write("jmp ", label1);
    write(label2, ":");
}


// doesn't modify the contetn of eax
void AssemblyGenerator::compileMult(Index index, Cell factor)
{
    // don't have to do anything on zero mult
    if (factor == 0) return;

    // just add
    if (factor == 1) {
        write("addb %al, ", index, ptr());
    }
    else if (factor == 255) {
        write("subb %al, ", index, ptr());
    }
    // check if factor is power of two
    else if ((factor & (factor - 1)) == 0) {
        if (factor == 2) {
            write("leal (%rax,%rax), %ecx");
            write("addb %cl, ", index, ptr());
        }
        else if (factor <= 8) {
            write("leal (,%eax,", int(factor), "), %ecx");
            write("addb %cl, ", index, ptr());
        }
        else {
            int tz = trailingZeroes(factor);
            write("movl %eax, %ecx");
            write("shl $", tz, ", %ecx");
            write("addb %cl, ", index, ptr());
        }
    }
    else {
        write("imull $", int(factor), ", %eax, %ecx");
        write("addb %cl, ", index, ptr());
    }
}

int AssemblyGenerator::trailingZeroes(unsigned long long x) {
#if __GNUC__ > 4 && 0
    return __builtin_ctzll(x);
#else
    for (int i = 0; i < 64; i++)
        if (x & (1ULL << i))
            return i;
    return 0;
#endif
}